#
#    Copyright 2010 Chandler "Omnirizon" Armstrong
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.




######################################################################
from __future__ import division
import re
from collections import defaultdict, deque
from pygame.font import Font
from pygame.rect import Rect
from pygame.sprite import Sprite
from pygame import Surface
from pygame.transform import scale
from pygame.locals import *




######################################################################
# constants
SPECIALS = ['/', '{', '}']                                            # special characters
WHITESPACE = {'n' : '\n'}




######################################################################
# globals
Macros = {}




######################################################################
# private functions
def _iswhitespace(char):
    whitespace = re.compile('\s+')                                        # re to detect whitespace
    if isinstance(char, str):
        assert len(char) == 1
        if whitespace.search(char): return True
        return False
    return False



def _scale_token(token, (width, height)):
    Token = _Token

    links, token_iswhitespace, token_str = token.links, token.iswhitespace, token.str
    surfs = [scale(token, (width, height))]

    ratio_w = width / token.get_width()
    ratio_h = height / token.get_height()

    # given token height, modify link rect y
    shift_x = 0
    for rect in [rect for v in links.values() for rect in v]:
        rect_w = rect.w
        rect.w *= ratio_w
        rect.h *= ratio_h
        rect.x += shift_x
        rect.y += (height - rect.h)    
        shift_x = rect.w - rect_w

    return Token((width, height), links, surfs, token_iswhitespace, token_str)



def _token_builder(interpreted_txt):
    iswhitespace = _iswhitespace
    Token = _Token

    links = defaultdict(list)

    token_iswhitespace = True

    surfs, x, y = [], 0, 0
    for (envs, chars) in interpreted_txt:
        bkg, color, font = envs['bkg'], envs['color'], envs['font']
        strbuff, surfbuff = [], []
        for char in chars:
            if not iswhitespace(char): token_iswhitespace = False

            if isinstance(char, str):
                if iswhitespace(char): char = ' '
                strbuff.append(char)

            else:
                if strbuff: surfbuff.append(font.render(''.join(strbuff), 0, color, bkg))
                surfbuff.append(char)
                strbuff = []

        if strbuff: surfbuff.append(font.render(''.join(strbuff), 0, color, bkg))

        if surfbuff:
            # calculate link rects
            link = envs['link']
            surfbuff_w = sum(surf.get_width() for surf in surfbuff)
            surfbuff_h = max(surf.get_height() for surf in surfbuff)
            links[link].append(Rect(x, 0, surfbuff_w, surfbuff_h))
            x += surfbuff_w
            # extend surfbuff to surfs and reset surfbuff
            surfs.extend(surfbuff)
            surfbuff = []

    # get token width and height
    width = sum(surf.get_width() for surf in surfs)
    height = max(surf.get_height() for surf in surfs)

    # given token height, modify link rect y
    for rect in [rect for v in links.values() for rect in v]: rect.y += (height - rect.h)

    token_str = ''.join(str(char) for (envs, chars) in interpreted_txt for char in chars)

    return Token((width, height), links, surfs, token_iswhitespace, token_str)




######################################################################
# private classes
class _Token(Surface):


    def __init__(self, (width, height), links, surfs, iswhitespace, token_str):
        self.links = links
        self.iswhitespace = iswhitespace
        self.str = token_str

        # construct Surface superclass
        Surface.__init__(self, (width, height))

        # blit each token onto self
        x = 0
        for surf in surfs:
            w, h = surf.get_size()
            y = height - h
            self.blit(surf, (x, y))
            x += w


    def __str__(self):
        return self.str



class _Line(Surface):
# Line class


    def __init__(self, line, surf_w, justify):
    # renders a line (list of tokens)
    # accepts a line (list of tokens), width of line, and justification
    # returns Surface object with the tokens justified upon it given the justify argument
        self.line = line
        self.links = defaultdict(list)

        links = self.links
        scale_token = _scale_token

        # get the line width, line height, whitespace used in the line, and freespace remaining in the line
        line_w = sum(token.get_width() for token in line)             #
        line_h = max(token.get_height() for token in line)            #
        total_whitespace = sum(token.get_width() for token in line if token.iswhitespace)#
        freespace = surf_w - line_w                                   # set the freespace available in the surface

        # initialize surface
        Surface.__init__(self, (surf_w, line_h))                      #

        # set x
        if justify == 'right': x = freespace                          # if right justified: x = freespace
        elif justify == 'center': x = freespace / 2                   # if center justified: x = freespace divided by 2
        else: x = 0                                                   # else (is left justified): x = 0

        # if justified: modify whitespace widths
        if justify == 'justified':                                    #
            stretch = 1.0 + (freespace / total_whitespace)              # the amount to stretch whitespace
            sub_w = 0
            for i, token in enumerate(line):                            # for i, token in enumerate tokens
                if token.iswhitespace:
                    token_w = token.get_width() * stretch
                    scale_w = int(token_w)
                    sub_w += token_w % scale_w
                    while sub_w > 1:
                        sub_w -= 1
                        scale_w += 1
                    line[i] = scale_token(token, (scale_w, token.get_height()))# if token is whitespace: replace token with a scaled token

        for token in line:                                            # for token in line:
            w, h = token.get_size()                                     # cache token width, height
            y = (line_h - h)                                            # get token y
            self.blit(token, (x, y))                                    # blit token to surface at x, y

            for link in token.links:                                    # for link in token
                for rect in token.links[link]:                            # for rect in link
                    links[link].append(rect.move(x, y))                     # move rect to token's position in line and append to links

            x += w                                                      # update x with object width


    def __str__(self):
        return ''.join(str(token) for token in self.line)



######################################################################
# public classes
class Main(Sprite):
#   Main glyph class


    ##################################################################
    # class methods
    def __init__(self, rect, **kwargs):
    #   Initialize a glyph object
        self.image = Surface(rect.size)
        self.rect = rect
        self._dest = Rect(0, 0, 0, 0)                                 # initial destination rect for blitting to viewing surface
        self._envs = [('bkg', kwargs['bkg']),                         # background color
                      ('color', kwargs['color']),                     # font color
                      ('font', kwargs['font']),                       # font
                      ('link', None)]                                 # link
        self.spacing = kwargs['spacing']                              # line spacing

        self._buff = deque()                                          # rendered text buffer
        self.links = defaultdict(list)                                # dict of (link, [rects]) pairs


    ##################################################################
    # helper methods
    def __read_env(self, _txt_):
    # helper function to interpret
    # accepts the _txt_ iterable at an environment starting point
    # returns (environment type, environment) tuple (e.g (font, Font object))
        r = re.compile('(\w+)(?:\s+([\w\s\.,\\\\/]+))?\s*;')          # env_name arguments (arguments may be paths and contain \ or /)
        charbuffer = ''                                               # initialize character buffer
        for i, char in _txt_:                                         # iterate over _txt_
            charbuffer += char                                          # add char to charbuffer
            s = r.search(charbuffer)                                    # search charbuffer using r
            if s:                                                       # if search successful
                env = s.groups()                                          # get environment
                if env[0] in Macros: return Macros[env[0]]                # if environment in Macros: return the Macro's environment
                # new environment types must be added here

                elif env[0] == 'bkg': return ('bkg',  tuple([int(e) for e in env[1].split(',')]))# if a background environment: return new color

                elif env[0] == 'color': return ('color',  tuple([int(e) for e in env[1].split(',')]))# if a color environment: return new color

                elif env[0] == 'font':                                    # if a font environment:
                    path, size = env[1].split(',')                          # get the font location and size
                    return ('font', Font(path, int(size)))                  # return new font

                elif env[0] == 'link': return ('link', env[1].strip())    # if a link environment: return new link

                else: raise ValueError, env[0] + ' is an unrecognized environment'# else: raise ValueError, the environment is not recognized


    # the space func could take a size and return a surface or rect...  really only need to shift tokens.  surfs do that without requirng tokens to have rects
    def __read_func(self, _txt_):
    # helper function to interpret
    # accepts the _txt_ iterable at function starting point
    # returns (function type, function results) tuple (eg. (space, Surface obejct))
        r = re.compile('(\w+){(.*?)}')
        i, char = _txt_.next()

        if char in SPECIALS: return char
        if char in WHITESPACE: return WHITESPACE[char]

        charbuff = char
        for i, char in _txt_:
            charbuff += char
            s = r.search(charbuff)
            if s:
                func, args = s.groups()
                if func in Macros: return Macros[func]

                if func == 'space': return Surface((int(args), 1))

                if charbuff == 'img': return image.load(args).convert()


    ##################################################################
    # private methods
    def _interpret(self, txt):
    # iterprets glyph markup language
    # accepts string literal
    # returns a list of rendered text and functions
        envs = self._envs                                             # cache envs
        read_env, read_func = self.__read_env, self.__read_func       # cache read_env, read_func, render helper function
        iswhitespace = _iswhitespace                                  # cache iswhitespace function

        charbuff, interpreted_txt, prevchar = [], [], ''              # initialize charbuff, renderbuff, interpreted text, and previous character
        _txt_ = enumerate(txt)                                        # initialize enumerated txt iterator
        for i, char in _txt_:                                         # iterate over text
            if iswhitespace(char) and iswhitespace(prevchar): continue  # if char and previous char are whitespace: continue

            elif char == '/':                                           # if a function:
                char = read_func(_txt_)                                   # read the function
                charbuff.append(char)                                     # append function value to charbuff
                prevchar = char                                           # set prevchar to function value

            elif char == '{':                                           # elif a left-curly bracket: a new environment has started
                # using dict(envs) allows new environments to overwrite default environments, which are in the beginning of the list
                interpreted_txt.append((dict(envs), charbuff))            # append (dict(envs), charbuff) to interpreted text
                charbuff = []                                             # reset charbuffer
                envs.append(read_env(_txt_))                              # append new environemt

            elif char == '}':                                           # elif a right-curly bracket: an environment has ended
                interpreted_txt.append((dict(envs), charbuff))            # append (dict(envs), charbuff) to interpreted text
                charbuff = []                                             # reset charbuffer
                envs.pop()                                                # pop last environment

            else:                                                       # else is a normal character:
                charbuff.append(char)                                     # append chacter to character buffer
                prevchar = char                                           # previous character = current character

        if charbuff: interpreted_txt.append((dict(envs), charbuff))   # if any characters remain in buffer: append (dict(envs), charbuff) to interpreted text

        return interpreted_txt                                        # return interpreted text


    def _tokenize(self, interpreted_txt):
    # tokenizes text
    # accepts (envs, charbuff)
    # returns a list of tokens
        iswhitespace, token_builder = _iswhitespace, _token_builder

        charbuff, _interpreted_txt, tokenized_txt = [], [], []
        for (envs, chars) in interpreted_txt:
            for char in chars:
                if iswhitespace(char):
                    if charbuff:
                        _interpreted_txt.append((envs, charbuff))
                        charbuff = []

                    if _interpreted_txt:
                        yield token_builder(_interpreted_txt)
                        _interpreted_txt = []

                    yield token_builder([(envs, [char])])

                else: charbuff.append(char)

            if charbuff:
                _interpreted_txt.append((envs, charbuff))
                charbuff = []

        if _interpreted_txt: yield token_builder(_interpreted_txt)


    def _wrap(self, tokenized_txt, justify):
    #   wrap interpreted text to page width
    #   accepts a list of Token objects tuples
    #   returns a list of Line objects, each wrapped to self.rect
        Line = _Line
        rect_w = self.rect.w                                          # cache rect width

        # linesize tracks the current size of the rendered line since moving between environments mean there will be multiple surfaces that need to be glued together
        line = []                                                     # initialize line, and linesize
        for token in tokenized_txt:                                   # iterate over tokenized text
            token_w = token.get_width()
            if token_w > rect_w:                                        # if size of token > rect width:
                raise ValueError("The token " + token + " is "            # raise ValueError( token too big to fit in rect width )
                                 + str((token_w - rect_w))
                                 + " pixels too wide to fit in the rect passed.")

            # the token fits, process it and check if the line still fits inside rect area, if not, append line without token, reinitialize line with token
            line.append(token)                                          # append token to line
            if str(token) == '\n':                                      # if token == newline
                if justify == 'justified': _justify = 'left'              # if justified: set justify to 'left' (do not want to justify a line that would not wrap)
                else: _justify = justify                                  # else: pass
                yield Line(line, rect_w, _justify)                        # yield current line
                line = []                                                 # reset line

            elif sum(token.get_width() for token in line) > rect_w:     # if size of line > rect width
                token = line.pop()                                        # pop last token from line
                if line[-1].iswhitespace: line = line[:-1]                # remove single trailing whitespace
                yield Line(line, rect_w, justify)                         # yield wrapped line
                line = []                                                 # reinitialize _line, line, linesize, and number of objects
                if not token.iswhitespace: line.append(token)             # if the token is not whitespace: append token as first token of new line

        if line:                                                      #
            if justify == 'justified': _justify = 'left'                #
            else: _justify = justify                                    #
            yield Line(line, rect_w, _justify)                          # if line remaining: yield line


    ##################################################################
    # public methods
    def input(self, txt, justify = None):
    # interprets, renders, wraps, and justifies input text
    # accepts raw text and a justify command
    # returns nothing
        buff, interpret, tokenize, wrap = self._buff, self._interpret, self._tokenize, self._wrap
        interpreted_txt = interpret(txt)                              #
        tokens = tokenize(interpreted_txt)                            #
        lines = wrap(tokens, justify)                                 #
        buff.extend(lines)                                            #


    def update(self):
    # updates the surface with the text in the buffer, deletes buffer
    # accepts nothing
    # returns nothing
        buff, dest, image, links, rect, spacing = self._buff, self._dest, self.image, self.links, self.rect, self.spacing
        while buff:                                                   #
            line = buff.popleft()                                       #
            line_h = line.get_height()                                  #
            if dest.y + line_h > rect.h:                                #
                buff.append(line)                                         #
                break                                                     #
            else:                                                       #
                image.blit(line, dest)                                    #

                for link in line.links:                                   # for link in token
                    for _rect in line.links[link]:                          # for rect in link
                        links[link].append(_rect.move(dest.topleft))          # move rect to token's position in line and append to links

                dest.y += line_h + spacing                                #


    def clear(self, surface_dest, background):
    # draws background over surface_dest using self.rect and resets self._dest
    # accepts surface_dest: the destination surface that self.image has been drawn to
    #         background: the background to draw over the areas that self.image was drawn to on surface_dest
    # returns nothing
        rect = self.rect
        self.image = Surface(rect.size)
        self._dest = Rect(0, 0, 0, 0)
        self.links = defaultdict(list)
        surface_dest.blit(background, rect, rect)


    def get_collisions(self, mpos):
        links, rect = self.links, self.rect
        for link in links:
            for _rect in links[link]:
                if _rect.move(rect.topleft).collidepoint(mpos): return link
"""



    def scroll(self):
    # scrolls the surface and destination rect to accomodate text in the buffer before updating
    # accepts nothing
    # returns nothing
        dest, image, spacing = self._dest, self.image, self.spacing
        h = sum(line.get_height() + spacing for line in self.buffer)
        dest.move_ip(0, -h)
        image.scroll(0, -h)
        update()
"""
